@@ -1,6 +1,60 @@
+-- |
+-- Module:      Test.Tasty.Ingredients.Rerun
+-- Copyright:   Oliver Charles (c) 2014, Andrew Lelechenko (c) 2019
+-- Licence:     BSD3
+--
+-- This ingredient
+-- for <https://hackage.haskell.org/package/tasty tasty> testing framework
+-- allows to filter a test tree depending
+-- on an outcome of the previous run.
+-- This may be useful in many scenarios,
+-- especially when a test suite grows large.
+--
+-- The behaviour is controlled by command-line options:
+--
+-- * @--rerun@ @ @
+--
+--     Rerun only tests, which failed during the last run.
+--     If the last run was successful, execute a full test
+--     suite afresh. A shortcut for @--rerun-update@
+--     @--rerun-filter failures,exceptions@
+--     @--rerun-all-on-success@.
+--
+-- * @--rerun-update@ @ @
+--
+--     Update the log file to reflect latest test outcomes.
+--
+-- * @--rerun-filter@ @CATEGORIES@
+--
+--     Read the log file and rerun only tests from a given
+--     comma-separated list of categories: @failures@,
+--     @exceptions@, @new@, @successful@. If this option is
+--     omitted or the log file is missing, rerun everything.
+--
+-- * @--rerun-all-on-success@ @ @
+--
+--     If according to the log file and @--rerun-filter@ there
+--     is nothing left to rerun, run all tests. This comes
+--     especially handy in @stack test --file-watch@ or
+--     @ghcid@ scenarios.
+--
+-- * @--rerun-log-file@ @FILE@
+--
+--     Location of the log file (default: @.tasty-rerun-log@).
+--
+-- To add it to your test suite just replace
+-- 'Tasty.defaultMain' with
+-- 'defaultMainWithRerun' or wrap arguments
+-- of 'Tasty.defaultMainWithIngredients'
+-- into 'rerunningTests'.
+
 {-# LANGUAGE DeriveDataTypeable #-}
 {-# LANGUAGE FlexibleContexts #-}
-module Test.Tasty.Ingredients.Rerun (rerunningTests) where
+
+module Test.Tasty.Ingredients.Rerun
+  ( defaultMainWithRerun
+  , rerunningTests
+  ) where
 
 import Prelude hiding (filter)
 
@@ -8,14 +62,13 @@ import Control.Applicative
 import Control.Arrow ((>>>))
 import Control.Monad (when)
 import Control.Monad.Trans.Class (lift)
-import Data.Char (isSpace)
+import Data.Char (isSpace, toLower)
 import Data.Foldable (asum)
+import Data.List (intercalate)
 import Data.List.Split (endBy)
 import Data.Maybe (fromMaybe)
-import Data.Monoid (mconcat)
+import Data.Monoid (Any(..), mempty)
 import Data.Proxy (Proxy(..))
-import Data.Semigroup.Applicative (Traversal(..))
-import Data.Tagged (Tagged(..), untag)
 import Data.Typeable (Typeable)
 import System.IO.Error (catchIOError, isDoesNotExistError)
 
@@ -34,65 +87,74 @@ newtype RerunLogFile = RerunLogFile FilePath
   deriving (Typeable)
 
 instance Tasty.IsOption RerunLogFile where
-  optionName = Tagged "rerun-log-file"
-  optionHelp = Tagged "The path to which rerun's state file should be saved"
+  optionName = return "rerun-log-file"
+  optionHelp = return "Location of the log file (default: .tasty-rerun-log)"
   defaultValue = RerunLogFile ".tasty-rerun-log"
   parseValue = Just . RerunLogFile
-
+  optionCLParser = Tasty.mkOptionCLParser (OptParse.metavar "FILE")
 
 --------------------------------------------------------------------------------
 newtype UpdateLog = UpdateLog Bool
   deriving (Typeable)
 
 instance Tasty.IsOption UpdateLog where
-  optionName = Tagged "rerun-update"
-
-  optionHelp = Tagged "If present the log file will be updated, otherwise it \
-                      \will be left unchanged"
-
+  optionName = return "rerun-update"
+  optionHelp = return "Update the log file to reflect latest test outcomes"
   defaultValue = UpdateLog False
-
-  parseValue = Just . UpdateLog . const True
-
-  optionCLParser = fmap UpdateLog $ OptParse.switch $ mconcat
-    [ OptParse.long name
-    , OptParse.help helpString
-    ]
-    where
-    name = untag (Tasty.optionName :: Tagged UpdateLog String)
-    helpString = untag (Tasty.optionHelp :: Tagged UpdateLog String)
+  parseValue = fmap UpdateLog . Tasty.safeReadBool
+  optionCLParser = Tasty.mkFlagCLParser mempty (UpdateLog True)
 
 --------------------------------------------------------------------------------
 data Filter = Failures | Exceptions | New | Successful
-  deriving (Eq, Ord)
+  deriving (Eq, Ord, Enum, Bounded, Show)
 
 parseFilter :: String -> Maybe Filter
-parseFilter "failures" = Just Failures
-parseFilter "exceptions" = Just Exceptions
-parseFilter "new" = Just New
-parseFilter "successful" = Just Successful
-parseFilter _ = Nothing
+parseFilter s = lookup s (map (\x -> (map toLower (show x), x)) everything)
 
 --------------------------------------------------------------------------------
 everything :: [Filter]
-everything = [Failures, Exceptions, New, Successful]
+everything = [minBound..maxBound]
 
 --------------------------------------------------------------------------------
 newtype FilterOption = FilterOption (Set.Set Filter)
   deriving (Typeable)
 
 instance Tasty.IsOption FilterOption where
-  optionName = Tagged "rerun-filter"
-
-  optionHelp = Tagged "A comma separated list to specify which tests to run when\
-                      \ comparing against previous test runs. Valid options \
-                      \are: everything, failures, exceptions, new"
-
+  optionName = return "rerun-filter"
+  optionHelp = return
+    $  "Read the log file and rerun only tests from a given comma-separated list of categories: "
+    ++ map toLower (intercalate ", " (map show everything))
+    ++ ". If this option is omitted or the log file is missing, rerun everything."
   defaultValue = FilterOption (Set.fromList everything)
-
   parseValue =
     fmap (FilterOption . Set.fromList) . mapM (parseFilter . trim) . endBy ","
     where trim = reverse . dropWhile isSpace . reverse . dropWhile isSpace
+  optionCLParser = Tasty.mkOptionCLParser (OptParse.metavar "CATEGORIES")
+
+--------------------------------------------------------------------------------
+newtype AllOnSuccess = AllOnSuccess Bool
+  deriving (Typeable)
+
+instance Tasty.IsOption AllOnSuccess where
+  optionName = return "rerun-all-on-success"
+  optionHelp = return "If according to the log file and --rerun-filter there is nothing left to rerun, run all tests. This comes especially handy in `stack test --file-watch` or `ghcid` scenarios."
+  defaultValue = AllOnSuccess False
+  parseValue = fmap AllOnSuccess . Tasty.safeReadBool
+  optionCLParser = Tasty.mkFlagCLParser mempty (AllOnSuccess True)
+
+--------------------------------------------------------------------------------
+newtype Rerun = Rerun { unRerun :: Bool }
+  deriving (Typeable)
+
+instance Tasty.IsOption Rerun where
+  optionName = return "rerun"
+  optionHelp = return "Rerun only tests, which failed during the last run. If the last run was successful, execute a full test suite afresh. A shortcut for --rerun-update --rerun-filter failures,exceptions --rerun-all-on-success"
+  defaultValue = Rerun False
+  parseValue = fmap Rerun . Tasty.safeReadBool
+  optionCLParser = Tasty.mkFlagCLParser mempty (Rerun True)
+
+rerunMeaning :: (UpdateLog, AllOnSuccess, FilterOption)
+rerunMeaning = (UpdateLog True, AllOnSuccess True, FilterOption (Set.fromList [Failures, Exceptions]))
 
 --------------------------------------------------------------------------------
 data TestResult = Completed Bool | ThrewException
@@ -100,24 +162,51 @@ data TestResult = Completed Bool | ThrewException
 
 
 --------------------------------------------------------------------------------
--- | This 'Tasty.Ingredient' transformer adds various @--rerun@ options to your
--- test program. These flags add stateful execution of your test suite, allowing
--- you to rerun only tests that are failing from the previous run, or tests that
--- that have been added since the last test ran, once the 'Tasty.TestTree' has
--- been filtered.
+
+-- | Drop-in replacement for 'Tasty.defaultMain'.
 --
--- The input list of 'Tasty.Ingredient's specifies the 'Tasty.Ingredients's that
--- will actually work with the filtered 'Tasty.TestTree'. Normally, you'll want
--- at least 'Tasty.Test.Runners.consoleTestReporter'.
+-- > import Test.Tasty
+-- > import Test.Tasty.Ingredients.Rerun
+-- >
+-- > main :: IO ()
+-- > main = defaultMainWithRerun tests
+-- >
+-- > tests :: TestTree
+-- > tests = undefined
+defaultMainWithRerun :: Tasty.TestTree -> IO ()
+defaultMainWithRerun =
+  Tasty.defaultMainWithIngredients
+    [ rerunningTests [ Tasty.listingTests, Tasty.consoleTestReporter ] ]
+
+-- | Ingredient transformer, to use with
+-- 'Tasty.defaultMainWithIngredients'.
+--
+-- > import Test.Tasty
+-- > import Test.Tasty.Runners
+-- > import Test.Tasty.Ingredients.Rerun
+-- >
+-- > main :: IO ()
+-- > main =
+-- >   defaultMainWithIngredients
+-- >     [ rerunningTests [ listingTests, consoleTestReporter ] ]
+-- >     tests
+-- >
+-- > tests :: TestTree
+-- > tests = undefined
 rerunningTests :: [Tasty.Ingredient] -> Tasty.Ingredient
 rerunningTests ingredients =
-  Tasty.TestManager (rerunOptions ++ existingOptions) $
+  Tasty.TestManager (rerunOptions ++ Tasty.ingredientsOptions ingredients) $
     \options testTree -> Just $ do
       let RerunLogFile stateFile = Tasty.lookupOption options
-          UpdateLog updateLog = Tasty.lookupOption options
-          FilterOption filter = Tasty.lookupOption options
+          (UpdateLog updateLog, AllOnSuccess allOnSuccess, FilterOption filter)
+            | unRerun (Tasty.lookupOption options) = rerunMeaning
+            | otherwise = (Tasty.lookupOption options, Tasty.lookupOption options, Tasty.lookupOption options)
 
-      filteredTestTree <- maybe testTree (filterTestTree testTree filter)
+      let nonEmptyFold = Tasty.trivialFold { Tasty.foldSingle = \_ _ _ -> Any True }
+          nullTestTree = not . getAny . Tasty.foldTestTree nonEmptyFold options
+          recoverFromEmpty t = if allOnSuccess && nullTestTree t then testTree else t
+
+      filteredTestTree <- maybe testTree (recoverFromEmpty . filterTestTree testTree filter)
                            <$> tryLoadStateFrom stateFile
 
       let tryAndRun (Tasty.TestReporter _ f) = do
@@ -132,7 +221,7 @@ rerunningTests ingredients =
                     fmap getConst $
                     flip State.evalStateT 0 $
                     Functor.getCompose $
-                    getTraversal $
+                    Tasty.getTraversal $
                     Tasty.foldTestTree (observeResults statusMap)
                                         options filteredTestTree
 
@@ -151,17 +240,15 @@ rerunningTests ingredients =
         -- simply run the above constructed IO action.
         Just e -> e
   where
-  existingOptions = flip concatMap ingredients $ \ingredient ->
-    case ingredient of
-      Tasty.TestReporter options _ -> options
-      Tasty.TestManager options _ -> options
-
-  rerunOptions = [ Tasty.Option (Proxy :: Proxy RerunLogFile)
+  rerunOptions = [ Tasty.Option (Proxy :: Proxy Rerun)
                  , Tasty.Option (Proxy :: Proxy UpdateLog)
                  , Tasty.Option (Proxy :: Proxy FilterOption)
+                 , Tasty.Option (Proxy :: Proxy AllOnSuccess)
+                 , Tasty.Option (Proxy :: Proxy RerunLogFile)
                  ]
 
   ------------------------------------------------------------------------------
+  filterTestTree :: Tasty.TestTree -> Set.Set Filter -> Map.Map [String] TestResult -> Tasty.TestTree
   filterTestTree testTree filter lastRecord =
     let go prefix (Tasty.SingleTest name t) =
           let requiredFilter = case Map.lookup (prefix ++ [name]) lastRecord of
@@ -190,6 +277,7 @@ rerunningTests ingredients =
 
     in go [] testTree
 
+  tryLoadStateFrom :: FilePath -> IO (Maybe (Map.Map [String] TestResult))
   tryLoadStateFrom filePath = do
     fileContents <- (Just <$> readFile filePath)
                       `catchIOError` (\e -> if isDoesNotExistError e
@@ -198,12 +286,16 @@ rerunningTests ingredients =
     return (read <$> fileContents)
 
   ------------------------------------------------------------------------------
+  saveStateTo :: FilePath -> IO (Map.Map [String] TestResult) -> IO ()
   saveStateTo filePath getTestResults =
     getTestResults >>= (show >>> writeFile filePath)
 
   ------------------------------------------------------------------------------
+  observeResults
+    :: IntMap.IntMap (STM.TVar Tasty.Status)
+    -> Tasty.TreeFold (Tasty.Traversal (Functor.Compose (State.StateT Int IO) (Const (Map.Map [String] TestResult))))
   observeResults statusMap =
-    let foldSingle _ name _ = Traversal $ Functor.Compose $ do
+    let foldSingle _ name _ = Tasty.Traversal $ Functor.Compose $ do
           i <- State.get
 
           status <- lift $ STM.atomically $ do
@@ -218,8 +310,8 @@ rerunningTests ingredients =
 
           Const (Map.singleton [name] status) <$ State.modify (+ 1)
 
-        foldGroup name children = Traversal $ Functor.Compose $ do
-          Const soFar <- Functor.getCompose $ getTraversal children
+        foldGroup name children = Tasty.Traversal $ Functor.Compose $ do
+          Const soFar <- Functor.getCompose $ Tasty.getTraversal children
           pure $ Const (Map.mapKeys (name :) soFar)
 
     in Tasty.trivialFold
